# encoding: utf-8

import logging
log = logging.getLogger("turbomail.dispatch")

from smtplib import SMTP


__all__ = ['Dispatch']


class Dispatch(object):

	def __init__(self, server, username=None, password=None, debug=False):
		self.server = server
		self.username = username
		self.password = password
		self.debug = debug
		self.log = None
		
		log.debug("Creating SMTP object.")
		self.connection = SMTP()
		self.connection.set_debuglevel(debug)

	def __del__(self):
		self.disconnect()
	
	@property
	def connected(self):
		return getattr(self.connection, 'sock', None) is not None

	def __nonzero__(self):
		return self.connected
	
	def connect(self):
		if self.connected: return
		
		log.info("Connecting to SMTP server %s." % self.server)
		self.connection.connect(self.server)

		try:
			self.connection.ehlo()
			self.connection.starttls()
			self.connection.ehlo()
			log.debug("TLS enabled on SMTP server.")

		except:
			log.warn("TLS not available on SMTP server.")

		if self.username and self.password:
			log.debug("Authenticating as %s." % self.username)
			self.connection.login(self.username, self.password)
	
	def disconnect(self):
		if not self.connected: return
		
		log.debug("Closing SMTP connection.")
		self.connection.quit()
	
	def __call__(self, message, bulk=False):
		self.connect()
		
		pack = message()
		
		log.info("\"%s\" >> \"%s\" :: \"%s\"" % (pack[0], pack[1], message.message['Subject']))
		self.connection.sendmail(*pack)
		
		if self.debug: self.disconnect()



# encoding: utf-8

import logging
log = logging.getLogger("turbomail.pool")

import math
from Queue import Queue, Empty
from threading import Event, Thread
from turbomail.dispatch import Dispatch


__all__ = ['Pool', 'MailPool']


class Pool(Thread):
	"""A threadpool which checks regularily for new jobs and spawns processes
	as needed.
	
	interval is the delay, in seconds, between spawn runs
	threads is the maximum number of threads running at any given time
	jobs is the maximum number of jobs a single thread is allowed to handle
	timeout is the amount of time a thread is willing to wait for new data
	"""
	
	def __init__(self, interval=10.0, threads=4, jobs=10, timeout=60, **kw):
		super(Pool, self).__init__()
		
		self._pool = 0
		self._queue = Queue()
		self._finished = Event()
		self._interval = interval
		self._threads = threads
		self._jobs = jobs
		self._timeout = timeout
		self._kw = kw
		
		log.debug("Thread pool created.")
	
	def enqueue(self, work):
		log.debug("Work enqueued.")
		self._queue.put(work)
	
	def shutdown(self):
		log.debug("Shutdown requested.")
		self._finished.set()

	def run(self):
		log.debug("Beginning thread pool main loop.")
		
		while True:
			if self._finished.isSet():
				log.debug("Shutdown request acknowledged.")
				break
			
			if not self._queue.empty():
				log.debug("Estimate %d work units in the queue." % self._queue.qsize())
				
			optimum_threads = min(self._threads, math.ceil(self._queue.qsize() / float(self._jobs)))
			
			if not self._queue.empty() and self._pool < optimum_threads:
				log.debug("Creating %d threads." % (optimum_threads - self._pool))
				for i in xrange(int(optimum_threads - self._pool)):
					thread = Thread(target=self.wrapper)
					thread.start()
					self._pool += 1
					
			self._finished.wait(self._interval)
			
		log.debug("Thread pool main loop has ended.")
	
	def wrapper(self):
		log.debug("Thread pool worker starting up.")
		
		self.worker()
		
		self._pool -= 1
		log.debug("Thread pool worker finished.")

	def worker(self):
		raise NotImplementedError


class MailPool(Pool):
	def worker(self):
		count = 0
		dispatch = Dispatch(**self._kw)
		
		log.debug("Worker starting work.")
		
		while True:
			if not count < self._jobs:
				log.debug("Worker death from old age.")
				break
			
			try:
				unit = self._queue.get(True, self._timeout)
				dispatch(unit)

			except Empty:
				log.debug("Worker death from starvation.")
				break
			
			count += 1			



# encoding: utf-8


__all__ = ['MailException', 'MailNotEnabledException', 'MailConfigurationException']


class MailException(Exception):
    pass

        
class MailNotEnabledException(MailException):
    def __str__(self):
        return "An attempt was made to use a facility of the Turbo-Mail " \
               "framework but outbound mail hasn't been enabled in the" \
               "config file [via mail.on]."
    
    
class MailConfigurationException(MailException):
    args = ()
    def __init__(self, message):
        self.message= message
        
    def __str__(self):
        return self.message



# encoding: utf-8

import turbogears

from email.Utils import formatdate
from email.MIMEMultipart import MIMEMultipart
from email.MIMEBase import MIMEBase
from email.MIMEText import MIMEText
from email.Encoders import encode_base64


__all__ = ['Message', 'TemplateMessage']


class Message(object):

	def __init__(self, sender, recipient, subject):
		self.message = MIMEMultipart() # ('related')
		self.plain = None
		self.rich = None
		self.attachments = []
		self.processed = False

		self.message['From'] = sender
		self.message['To'] = recipient
		self.message['Date'] = formatdate(localtime=True)
		self.message['Subject'] = subject
	
	def attach(self, file, name=None):
		part = MIMEBase('application', "octet-stream")

		fp = open(file, "rb")
		part.set_payload(fp.read())
		Encoders.encode_base64(part)

		if name:
			part.add_header('Content-Disposition', 'attachment; filename="%s"' % name)
		else:
			part.add_header('Content-Disposition', 'attachment; filename="%s"' % os.path.basename(file))
			
		self.attachments.append(part)
	
	def embed(self, file, name):
		from email.MIMEImage import MIMEImage
		
		fp = open(file, 'rb')
		part = MIMEImage(fp.read())
		fp.close()
		
		part.add_header('Content-ID', '<%s>' % name)
		
		self.attachments.append(part)

	def process(self):
		if self.plain and self.rich:
			alternative = MIMEMultipart('alternative')
			self.message.attach(alternative)
			
			alternative.attach(MIMEText(self.plain))
			alternative.attach(MIMEText(self.rich, 'html'))
		
		else:
			self.message.attach(MIMEText(self.plain))
		
		for attachment in self.attachments:
			self.message.attach(attachment)
		
		self.processed = True 
	
	def __call__(self):
		if not self.processed:
			self.process()
		
		return (self.message['From'], self.message['To'], self.message.as_string())


class TemplateMessage(Message):
	
	def __init__(self, sender, recipient, subject, template, variables):
		self.template = template
		self.variables = dict(sender=sender, recipient=recipient, subject=subject)
		self.variables.update(variables)
				
		super(TemplateMessage, self).__init__(sender, recipient, subject)
	
	def process(self):
		turbogears.view.base._load_engines()
		
		localvars = dict()
		
		for (i, j) in self.variables.iteritems():
			if callable(j):
				localvars[i] = j()
			else:
				localvars[i] = j
		
		render = turbogears.view.engines.get('kid').transform(localvars, self.template)
		render = turbogears.widgets.base.PlainHTML(elements=render)
		
		self.plain = render.serialize(output="plain")			
		self.rich = render.serialize(output="html")
		
		return super(TemplateMessage, self).process()



# encoding: utf-8

import logging
log = logging.getLogger("turbomail")

from turbomail.startup import start_extension, shutdown_extension
from turbomail.exceptions import *
from turbomail.message import Message, TemplateMessage


__all__ = ['MailException', 'MailNotEnabledException', 'MailConfigurationException', 'Message', 'TemplateMessage', 'enqueue']


queue = None
enqueue = None
